<?php
declare(strict_types=1);

/**
 * webEdition CMS
 *
 * $Rev: 9aed2d40b9fb $
 * $Author: Marc Kraemer <marc.kraemer@aldente-it.de> $
 * $Date: 2020-09-01 17:08 +0200 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */
abstract class we_editor_editFrameset{

	/**
	 * Searches for the first Page in the editor, which the user is allowed to see.
	 * If he can see the given Nr, then that page will be shown.
	 *
	 * @see     getFirstValidEditPageNr
	 *
	 * @param   we_contents_root $we_doc
	 * @param   int $EditPageNr

	 * @return   int
	 */
	private static function getFirstValidEditPageNr(we_contents_root $we_doc, int $EditPageNr): int{
		if($we_doc->isValidEditPage($EditPageNr) && we_base_permission::isUserAllowedForAction('switch_edit_page', $EditPageNr)){
			return $EditPageNr;
		}
		//	bugfix for new tag: we:hidePages
		foreach(array_keys($we_doc->EditPageNrs) as $key){
			//  the command in this case is swith_edit_page, because in this funtion
			//  the editor tries to select a certain edit_page
			//  in some cases it must switch it
			if(we_base_permission::isUserAllowedForAction('switch_edit_page', $we_doc->EditPageNrs[$key])){
				return $we_doc->EditPageNrs[$key];
			}
		}
		return -1;
	}

	private static function getTabs(we_contents_root $we_doc, int $predefined = we_base_constants::WE_EDITPAGE_PROPERTIES): int{
		// Check which tab the user can see
		return ($we_doc->mustCheckEditPageArray() ?
			self::getFirstValidEditPageNr($we_doc, $predefined) :
			$predefined);
	}

	private static function checkPerm(string $we_Table): void{
		switch($we_Table){
			case TEMPLATES_TABLE:
				$needPerm = we_base_perm::WORK_TEMPLATES_CAN_SEE;
				break;
			case FILE_TABLE:
				$needPerm = we_base_perm::WORK_DOCUMENTS_CAN_SEE;
				break;
			case VFILE_TABLE:
				$needPerm = we_base_perm::WORK_COLLECTIONS_CAN_SEE;
				break;
			case OBJECT_FILES_TABLE:
				$needPerm = we_base_perm::OBJECTFILES_CAN_SEE;
				break;
			case OBJECT_TABLE:
				$needPerm = we_base_perm::OBJECTS_CAN_SEE;
				break;
			default:
				$needPerm = '';
		}
		if($needPerm && !we_base_permission::hasPerm([$needPerm])){
			we_base_permission::noPermDialog(g_l('alert', '[no_perms_action]'));
		}
	}

	private static function checkWorkspacePerm(we_contents_root $we_doc, string $we_Table, string $we_ContentType): void{
		if(!$we_doc->ID){
			return;
		}
		if(($ws = we_users_util::get_ws($we_Table))){
			if(!(we_users_util::in_workspace($we_doc->ID, $ws, $we_Table))){
				switch($we_Table){
					case TEMPLATES_TABLE: //	different workspace. for template
						we_base_permission::noPermDialog(g_l('alert', '[' . (($we_ContentType === we_base_ContentTypes::FOLDER_TEMPLATE) ? we_base_ContentTypes::FOLDER_COMMON : $we_Table) . '][not_im_ws]'));
					case FILE_TABLE: //	only preview mode allowed for docs
						//	MUST change to Preview-Mode
						$_SESSION['weS']['EditPageNr'] = we_base_constants::WE_EDITPAGE_PREVIEW;
						break;
				}
			}
		}
		switch($we_doc->userHasAccess()){
			case we_contents_root::USER_HASACCESS:
				break;
			case we_contents_root::FILE_LOCKED:
			case we_contents_root::USER_NO_SAVE:
				$_SESSION['weS']['EditPageNr'] = we_base_constants::WE_EDITPAGE_PREVIEW;
				break;
			case we_contents_root::FILE_NOT_IN_USER_WORKSPACE:
			default:
				we_base_permission::noPermDialog(g_l('alert', '[no_perms]', false, true));
				break;
		}
	}

	private static function checkLocked(we_contents_root $we_doc): void{
		if($we_doc instanceof we_folder_objectFile && $we_doc->isClassFolder()){
			return;
		}

		$userID = $we_doc->isLockedByUser(); //	Check if file is locked.
		DB_WE::$inst->query('UPDATE ' . USER_TABLE . ' SET Ping=NULL WHERE Ping<(NOW()- INTERVAL ' . (we_base_constants::PING_TIME + we_base_constants::PING_TOLERANZ) . ' SECOND)');

		$filelocked = ($userID != 0 && $userID != we_users_user::getID());

		if(!$filelocked && // file can be edited
			//Lock the new file
			//	before lock - check if user can edit the file.
			$we_doc->userHasAccess() === we_contents_root::USER_HASACCESS && //	only when user has access to file
			(we_editor_functions::isNormalMode() ||
			$we_doc->EditPageNr != we_base_constants::WE_EDITPAGE_PREVIEW)){
			$we_doc->lockDocument();
		}

		if($we_doc instanceof we_contents_objectFile && (!$we_doc->canMakeNew())){ // at this time only in objectFiles
			we_base_permission::noPermDialog(g_l('alert', '[no_new][objectFile]'));
		}
	}

	private static function setOnload(we_contents_root $we_doc): string{
		// Don't do this with Templates and only in Preview Mode
		// in Edit-Mode all must be reloaded !!!
		// To remove this functionality - just use the second condition as well.
		return ($we_doc->ContentType !== we_base_ContentTypes::TEMPLATE ? 'setOnload();' : '');
	}

	public static function show(string $parastr = ''): void{
		$we_Table = we_base_request::_(we_base_request::TABLE, 'we_cmd', FILE_TABLE, 1);
		$we_ID = we_base_request::int('we_cmd', 0, 2);
		$we_ContentType = we_base_request::string('we_cmd', '', 3) ?: ($we_ID ? fS('SELECT ContentType FROM ' . DB_WE::$inst->escape($we_Table) . ' WHERE ID=' . $we_ID) : '');
		$we_transaction = we_base_WE::$transaction;

// init document
		try{
			$we_doc = we_contents_root::getSessionDocument($we_transaction);
		} catch (Exception $ex){
			$we_doc = we_document_base::initDoc($we_ContentType, $we_ID, $we_Table);
		}

		if(!$we_doc->fileExists){
			we_view_document::showNoResourceMsg();
			exit();
		}

		self::checkPerm($we_Table);

		$we_doc->InWebEdition = true;
//$we_doc->i_loadNavigationItems();
//	check template for hidePages
		$we_doc->setDocumentControlElements();

//	in SEEM-Mode the first page is the preview page.
//	when editing an image-document we go to edit page
		if(we_editor_functions::isSeeMode()){
			if($we_doc instanceof we_document_image){
				$_SESSION['weS']['EditPageNr'] = we_base_constants::WE_EDITPAGE_CONTENT;
				$we_doc->EditPageNr = we_base_constants::WE_EDITPAGE_CONTENT;
			} else {
				$_SESSION['weS']['EditPageNr'] = we_base_constants::WE_EDITPAGE_PREVIEW;
				$we_doc->EditPageNr = we_base_constants::WE_EDITPAGE_PREVIEW;
			}
		}

		if(($tid = we_base_request::int('we_cmd', -1, 10)) !== -1 && ($we_Table === FILE_TABLE) && ($we_ContentType === we_base_ContentTypes::WEDOCUMENT)){
			assert($we_doc instanceof we_document_webEdition);
			$we_doc->setTemplateID($tid);
			$_SESSION['weS']['EditPageNr'] = self::getTabs($we_doc, we_base_constants::WE_EDITPAGE_CONTENT);
		} elseif(!$we_ID){
			$_SESSION['weS']['EditPageNr'] = self::getTabs($we_doc, we_base_constants::WE_EDITPAGE_PROPERTIES);
		}

//predefine ParentPath
		if(we_base_request::string('we_cmd', '', 0) === 'new_document'){
			if(($pid = we_base_request::int('we_cmd', 0, 5)) && $we_doc->ParentID == 0){
				if(in_array($we_doc->ContentType, we_base_ContentTypes::FOLDER_TYPES)){
					assert($we_doc instanceof we_folder_base);
					$we_doc->setParentID($pid);
				}
			}
			if(($copyID = we_base_request::int('copyID', 0))){
				$we_doc->copyDoc($copyID);
			}
		}

		if(($doct = we_base_request::int('we_cmd', -1, 8)) !== -1 && ($we_Table === FILE_TABLE) && ($we_ContentType === we_base_ContentTypes::WEDOCUMENT)){
			assert($we_doc instanceof we_document_webEdition);
			$we_doc->changeDoctype($doct);
			$_SESSION['weS']['EditPageNr'] = self::getTabs($we_doc, we_base_constants::WE_EDITPAGE_CONTENT);
		} else if($doct !== -1 && ($we_Table === OBJECT_FILES_TABLE) && ($we_ContentType === we_base_ContentTypes::OBJECT_FILE)){
			assert($we_doc instanceof we_contents_objectFile);
			$we_doc->TableID = $doct;
			$we_doc->setRootDirID(true);
			$we_doc->restoreDefaults();
			$_SESSION['weS']['EditPageNr'] = self::getTabs($we_doc, we_base_constants::WE_EDITPAGE_CONTENT);
		}

		self::checkWorkspacePerm($we_doc, $we_Table, $we_ContentType);

		/* if(isset($we_sess_folderID) && is_array($we_sess_folderID) && (!$we_doc->ID) && $we_sess_folderID[$we_doc->Table]){
		  $we_doc->setParentID($we_sess_folderID[$we_doc->Table]);
		  } */

		if($we_doc->ID == 0){
			$we_doc->EditPageNr = self::getTabs($we_doc, we_base_constants::WE_EDITPAGE_PROPERTIES);
		} else if(isset($_SESSION['weS']['EditPageNr'])){
			if(we_modules_info::isActive(we_modules_info::SHOP)){
				$we_doc->checkTabs();
			}

			$we_doc->EditPageNr = (in_array($_SESSION['weS']['EditPageNr'], $we_doc->EditPageNrs) ?
				self::getTabs($we_doc, $_SESSION['weS']['EditPageNr']) :
				//	Here we must get the first valid EDIT_PAGE
				self::getFirstValidEditPageNr($we_doc, we_base_constants::WE_EDITPAGE_CONTENT));
		}

		if($we_doc->EditPageNr === -1){ //	there is no view available for this document
			//	show errorMessage - no view for this document (we:hidePages)
			echo we_html_tools::getHtmlTop('', '', we_html_element::body(['class' => 'weDialogBody'], we_html_tools::dialogLayout(we_html_tools::alertAttentionBox(g_l('alert', '[no_views][description]'), we_html_tools::TYPE_ALERT, 500, true), g_l('alert', '[no_views][headline]'))
				)
			);
			exit;
		}

		self::checkLocked($we_doc);

// objects need to know the last webEdition Path, because of Workspaces
		if($we_doc->ContentType === we_base_ContentTypes::WEDOCUMENT){
			$_SESSION['weS']['last_webEdition_document'] = ['Path' => $we_doc->Path];
		}

// get default code
		if(!$we_doc->getElementS('data')){
			$we_doc->setElementS('data', ($we_doc->ContentType === we_base_ContentTypes::TEMPLATE && ($cmd10 = we_base_request::string('we_cmd', '', 10)) ?
					(base64_decode($cmd10) ?: '' ) :
					we_base_ContentTypes::getDefaultCode($we_doc->ContentType))
			);
		}

		if(!$we_doc->ID && $we_doc instanceof we_document_binary){
			$we_doc->EditPageNr = we_base_constants::WE_EDITPAGE_CONTENT;
		}
		$we_doc->saveInSession($we_transaction);
		$fid = we_base_request::string('frameId');
		switch($_SESSION['weS']['we_mode']){
			case we_base_constants::MODE_SEE:
				$showContentEditor = true;
				$headerSize = 1;
				break;
			case we_base_constants::MODE_NORMAL:
			default:
				$showContentEditor = ($we_doc->EditPageNr === we_base_constants::WE_EDITPAGE_CONTENT && substr($we_doc->ContentType, 0, 5) === 'text/' && $we_doc->ContentType !== we_base_ContentTypes::WEDOCUMENT);
				$headerSize = 39;
		}

		echo we_html_tools::getHtmlTop('', we_html_element::jsScript(JS_DIR . 'we_edit_frameset.js', "edit_framesetStart('" . htmlspecialchars(addcslashes($we_doc->getText(), '\'') . "','" . $we_doc->Path . "','" . $we_doc->Table . "'," . $we_doc->ID . ",'" . $we_transaction . "','" . $we_doc->ContentType . "','" . $parastr . "');"), [
				'id' => 'loadVarEdit_frameset',
				'data-editFrameset' => setDynamicVar([
					'USERACCESS' => $we_doc->userHasAccess() == we_contents_root::USER_HASACCESS
			])]), we_html_element::body([
				'onload' => "_EditorFrame.initEditorFrameData({'EditorIsLoading': false});" . $we_doc->getFrameLoadCode(),
				'onunload' => "doUnload();",
				'class' => "editFrameset"
				],
//FIXME: if we want to remove these iframes, e.g. EditorFrameController.js enumerate the frames, make sure to get all
															 we_html_element::iframe('editHeader', we_contents_base::url(WEBEDITION_DIR . 'we_cmd.php?we_cmd[0]=we_gui_loadEditHeader'), 'height:' . $headerSize . 'px;', '', '', 'editHeader') .
				($showContentEditor ?
					we_html_element::iframe('editor_' . $fid, 'about:blank', 'display:none;top:' . $headerSize . 'px;', '', self::setOnload($we_doc), 'mainEditor') .
					we_html_element::iframe('contenteditor_' . $fid, we_contents_base::url(WEBEDITION_DIR . 'we_cmd.php?we_cmd[0]=load_editor') . ($parastr ? '&' . $parastr : '') . '&we_complete_request=1', 'top:' . $headerSize . 'px;', '', '', 'contenteditor') :
					we_html_element::iframe('editor_' . $fid, we_contents_base::url(WEBEDITION_DIR . 'we_cmd.php?we_cmd[0]=load_editor') . ($parastr ? '&' . $parastr : '') . '&we_complete_request=1', 'top:' . $headerSize . 'px;', '', self::setOnload($we_doc), 'mainEditor') .
					we_html_element::iframe('contenteditor_' . $fid, 'about:blank', 'display:none;top:' . $headerSize . 'px;', '', '', 'contenteditor')
				) .
				we_html_element::iframe('editFooter', we_contents_base::url(WEBEDITION_DIR . 'we_cmd.php?we_cmd[0]=we_gui_loadEditFooter'), '', '', '', 'editorButtonFrame')
		));
	}

}
