<?php
declare(strict_types=1);

/**
 * webEdition CMS
 *
 * $Rev: 33b8422b6be7 $
 * $Author: Marc Kraemer <marc.kraemer@aldente-it.de> $
 * $Date: 2020-10-08 17:46 +0200 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @category    webEdition
 * @license     http://www.gnu.org/copyleft/gpl.html  GPL
 */
abstract class we_object_seo{
	/** @var bool */
	private static $debug = false;

	private static function searchObj(string $urlLookingFor): array{
		/**
		 * url without query string
		 * we need this in some we_weTag_util::tag() and to avoid dublicate content
		 */
		$urlLookingFor = ltrim((URLENCODE_OBJECTSEOURLS ?
			strtr(urlencode($urlLookingFor), ['%2F' => '/', '//' => '/']) :
			strtr($urlLookingFor, ['//' => '/'])
			), '/');

		define('WE_REDIRECTED_SEO', $urlLookingFor);

		/**
		 * now, we looking for an object ID an starting with full path of the URL
		 * and then we are checking the URL from left to right
		 * e.g. /mainfolder/subfolder/part-1-of-seo-url/part-2-of-seo-url
		 * in /mainfolder/subfolder is a dynamic webEdition document aka Trigger document
		 * this Trigger document is part oh the whole URL but not of the object SEO-URL
		 *
		 * first check: mainfolder/subfolder/part-1-of-seo-url/part-2-of-seo-url --> nothing is found
		 * second check: subfolder/part-1-of-seo-url/part-2-of-seo-url --> nothing is found
		 * third check: part-1-of-seo-url/part-2-of-seo-url --> we get the object
		 * and so one
		 */
		//get all possible paths and have one query for it
		$allUrls = [];
		//maybe it is easier to have an extra hashed index. This can make the DB faster, since not the whole string has to be compared
		//$allHashes = [];
		while($urlLookingFor){
			$allUrls[] = $urlLookingFor;
			$allUrls[] = '/' . $urlLookingFor;
			//$allHashes[] = md5(strtolower($urlLookingFor));
			//reduce the rest of the given url and try again
			$urlLookingFor = ltrim(stristr($urlLookingFor, '/') ?: '', '/');
		}

		self::$debug && t_e('Search for urls', $allUrls);
		if(($object = getHash('SELECT `of`.ID,`of`.ParentID,`of`.Url,`of`.Workspaces,f.Path AS TriggerPath FROM ' . OBJECT_FILES_TABLE . ' `of` LEFT JOIN ' . FILE_TABLE . ' f ON (f.ID=`of`.TriggerID AND f.IsPublished=1 AND f.IsDynamic>0) WHERE `of`.IsPublished=1 AND `of`.Url IN("' . implode('","', array_map('escape_sql_query', $allUrls)) . '") ORDER BY LENGTH(`of`.Url) DESC LIMIT 1'))){
			/**
			 * we check if the given URL and DB Url are identical
			 * if not we redirect to the DB url to avoid dublicate content
			 */
			if(!in_array($object['Url'], $allUrls)){//different writing
				$pos = array_search(strtolower($object['Url']), array_map('strtolower', $allUrls));
				if($pos !== false){
					header('Location: ' . substr(WE_REDIRECTED_SEO, 0, -1 * strlen($allUrls[$pos])) . $object['Url'], true, 301);
					exit;
				}
			}
		}
		return $object;
	}

	private static function getTrigger(array $object): string{
		$checkForUrlMap = fS('SELECT Path FROM ' . FILE_TABLE . ' WHERE urlMap LIKE "' . $_SERVER['SERVER_NAME'] . '%" LIMIT 1'); //check for multi domains in webedition
		//cut the known seo-url from object of the whole URL
		$docPathOfUrl = $checkForUrlMap . we_base_file::addPreSlash(substr(WE_REDIRECTED_SEO, 0, strripos(WE_REDIRECTED_SEO, $object['Url']) ?: 0)); //@phan-suppress-current-line PhanParamSuspiciousOrder
		//get trigger document by url and/or (extra) workspaces by object properties
		$path = rtrim($docPathOfUrl, '/');
		$triggerDocPath = we_contents_objectFile::getNextDynDoc($path . DEFAULT_DYNAMIC_EXT, path_to_id($path), explode(',', $object['Workspaces']));
		self::$debug && t_e('doc path', $docPathOfUrl, 'Path of trigger', $triggerDocPath);


		if($triggerDocPath || !NAVIGATION_DIRECTORYINDEX_NAMES){
			return $triggerDocPath;
		}

//fallback
		//now we try to get trigger doc by the given SEO-URL and NAVIGATION_DIRECTORYINDEX_NAMES from preferences
		$dirIndexArray = array_map('escape_sql_query', array_map('trim', explode(',', NAVIGATION_DIRECTORYINDEX_NAMES)));
		if($dirIndexArray){
			$docPath = DB_WE::$inst->escape($docPathOfUrl);
			$docPath .= implode('","' . $docPath, $dirIndexArray);

			return fS('SELECT Path FROM ' . FILE_TABLE . ' WHERE IsPublished=1 AND IsDynamic=1 AND Path IN ("' . $docPath . '") LIMIT 1');
		}

		return '';
	}

	private static function updateRequest(string $triggerDocPath, array $object): void{
		//remove all cookies from Request String if set (if not, cookies are exposed on listviews etc & max interfer with given Cookies)
		$_REQUEST = array_merge($_GET, $_POST);

		//get query string if there
		$queryStringArray = [];
		$queryString = !empty($_SERVER['REDIRECT_QUERY_STRING']) ? $_SERVER['REDIRECT_QUERY_STRING'] :
			($triggerDocPath ?: //we use the default trigger document of object class
			($object['TriggerPath'] ?: ''));
		parse_str($queryString, $queryStringArray);

		//should we also send $_GET?
		$_GET = isset($_GET) ? array_merge($_GET, $queryStringArray) : $queryStringArray;

		$_REQUEST = empty($_REQUEST) ? $queryStringArray : array_merge($_REQUEST, $queryStringArray); //@phan-suppress-current-line PhanPluginSuspiciousParamOrderInternal
		$_REQUEST[WEOBJ_PARAM] = $_REQUEST['we_oid'] = $object['ID'];
	}

	public static function searchURL(): string{
		if(we_base_WE::$editmode){
			exit();
		}

		if(!we_modules_info::isActive(we_modules_info::OBJECT)){
			if(ERROR_DOCUMENT_NO_OBJECTFILE){
				we_html_tools::setHttpCode(SUPPRESS404CODE ? 200 : 404);
				return WEBEDITION_PATH . '../' . id2path(ERROR_DOCUMENT_NO_OBJECTFILE, FILE_TABLE);
			}
			exit();
		}

		$urlLookingFor = (string) (!empty($_SERVER['REDIRECT_URL']) && !strpos($_SERVER['REDIRECT_URL'], ltrim(WEBEDITION_DIR, '/')) ?
			urldecode($_SERVER['REDIRECT_URL']) :
			(!empty($_SERVER['REQUEST_URI']) && !strpos($_SERVER['REQUEST_URI'], ltrim(WEBEDITION_DIR, '/')) ?
			parse_url(urldecode($_SERVER['REQUEST_URI']), PHP_URL_PATH) :
			''));

		if(!$urlLookingFor){
			exit(2);
		}

		/**
		 * now we try to get the trigger document
		 */
		self::$debug && t_e('Search for url', $urlLookingFor);
		if(($object = self::searchObj($urlLookingFor))){
			self::$debug && t_e('Object found', $object);
			if(($triggerDocPath = self::getTrigger($object))){// now we have an object and a trigger document
				self::$debug && t_e('found trigger doc', $triggerDocPath);
				self::updateRequest($triggerDocPath, $object);

				$_SERVER['SCRIPT_FILENAME'] = WEBEDITION_PATH . '..' . $triggerDocPath;
				$_SERVER['SCRIPT_NAME'] = $_SERVER['SCRIPT_URL'] = $triggerDocPath;

				we_html_tools::setHttpCode(200);
				return $_SERVER['SCRIPT_FILENAME']; //trigger has a leading /
			}
		}

		/**
		 * nothing found show errorDoc404
		 */
		if(ERROR_DOCUMENT_NO_OBJECTFILE){
			we_html_tools::setHttpCode(SUPPRESS404CODE ? 200 : 404);
			if(FORCE404REDIRECT){
				header('Location: ' . id2path(ERROR_DOCUMENT_NO_OBJECTFILE, FILE_TABLE));
				exit;
			}
			return WEBEDITION_PATH . '../' . id2path(ERROR_DOCUMENT_NO_OBJECTFILE, FILE_TABLE);
		}
		exit(1);
	}

}
