<?php
declare(strict_types=1);

/**
 * RSS parser class.
 *
 * This class is a parser for Resource Description Framework (RDF) Site
 * Summary (RSS) documents. For more information on RSS see the
 * website of the RSS working group (http://www.purl.org/rss/).
 *
 * @author Martin Jansen <mj@php.net>
 * @version $Revision$
 * @access  public
 */
class we_xml_rss extends we_xml_browser{//use we_xml_browser
	/** @var array	 */
	private $insideTagStack = [];

	/** @var string	 */
	private $activeTag = '';

	/** @var array	 */
	private $channel = [];

	/** @var string */
	public $parseError = '';

	/** @var array	 */
	private $items = [];

	/** @var array	 */
	private $images = [];

	/** @var array<string,array> */
	private $fields = [
		'item' => [],
		'textinput' => [],
		'image' => [],
	];

	/** @var array	 */
	private $textinputs = [];

	/** @var array	 */
	private $attribs;

	/** @var array */
	private $last;

	/**
	 * @var array
	 */
	private const PARENT_TAGS = ['CHANNEL', 'ITEM', 'IMAGE', 'TEXTINPUT'];

	/**
	 * @var array
	 */
	private const CHANNEL_TAGS = ['TITLE', 'LINK', 'DESCRIPTION', 'IMAGE',
		'ITEMS', 'TEXTINPUT', 'LANGUAGE', 'COPYRIGHT',
		'MANAGINGEDITOR', 'WEBMASTER', 'PUBDATE', 'LASTBUILDDATE',
		'CATEGORY', 'GENERATOR', 'DOCS', 'CLOUD', 'TTL',
		'RATING'];

	/**
	 * @var array
	 */
	private const ITEM_TAGS = ['TITLE', 'LINK', 'DESCRIPTION', 'PUBDATE', 'AUTHOR', 'CATEGORY',
		'COMMENTS', 'ENCLOSURE', 'GUID', 'PUBDATE', 'SOURCE',
		'CONTENT:ENCODED'];

	/** @var array	 */
	private const IMAGE_TAGS = ['TITLE', 'URL', 'LINK', 'WIDTH', 'HEIGHT'];

	/** @var array	 */
	private const TEXTINPUT_TAGS = ['TITLE', 'DESCRIPTION', 'NAME', 'LINK'];

	/** @var array  List of allowed module tags
	 * Currently Dublin Core Metadata, blogChannel RSS module, CreativeCommons,
	 * Content and Syndication are supported.
	 */
	private const MODULE_TAGS = [
		'DC:TITLE', 'DC:CREATOR', 'DC:SUBJECT', 'DC:DESCRIPTION',
		'DC:PUBLISHER', 'DC:CONTRIBUTOR', 'DC:DATE', 'DC:TYPE',
		'DC:FORMAT', 'DC:IDENTIFIER', 'DC:SOURCE', 'DC:LANGUAGE',
		'DC:RELATION', 'DC:COVERAGE', 'DC:RIGHTS',
		'BLOGCHANNEL:BLOGROLL', 'BLOGCHANNEL:MYSUBSCRIPTIONS',
		'BLOGCHANNEL:MYSUBSCRIPTIONS', 'BLOGCHANNEL:CHANGES',
		'CC:LICENSE', 'CONTENT:ENCODED',
		'SY:UPDATEPERIOD', 'SY:UPDATEFREQUENCY', 'SY:UPDATEBASE',
	];

	/**
	 * @access public
	 * @param string $data File pointer, name of the RSS file, or an RSS string.
	 * @param string $charset Target charset encoding, use null (default) to use
	 *                default encoding (ISO-8859-1)
	 * @throws Exception
	 */
	protected function parseXML(string $data, string $charset = 'ISO-8859-1'): bool{
		if($data == ''){
			return false;
		}

		$parser = xml_parser_create();
		xml_parser_set_option($parser, XML_OPTION_TARGET_ENCODING, (strstr($charset, 'ISO-88') ? 'ISO-8859-1' : $charset));

		xml_set_object($parser, $this);
		xml_set_element_handler($parser, [$this, 'startHandler'], [$this, 'endHandler']);
		xml_set_character_data_handler($parser, [$this, 'cdataHandler']);
		//xml_set_default_handler($parser, 'defaultHandler');


		if(!xml_parse($parser, $data, true)){
			$this->parseError = xml_get_current_line_number($parser) . ': ' . xml_Error_string(xml_get_error_code($parser) ?: 0) . ' - ' .
				xml_get_current_byte_index($parser) . '(' . substr($data, (int) xml_get_current_byte_index($parser) - 10, 20) . ')';
			throw new Exception($data);
		}
		xml_parser_free($parser);
		return true;
	}

	/**
	 * Start element handler for XML parser
	 *
	 * @access private
	 * @param  resource $parser XML parser object
	 * @param  string $element XML element
	 * @param  array $attribs Attributes of XML tag
	 */
	function startHandler($parser, string $element, array $attribs): void{//@phan-suppress-current-line PhanUnusedPublicNoOverrideMethodParameter
		$element = str_replace('RSS:', '', $element);

		if(in_array($element, self::PARENT_TAGS, true)){
			array_push($this->insideTagStack, $element);
		} else {
			switch($element){
				case 'ENCLOSURE' :
					$this->attribs = $attribs;
					break;

				default:
					$this->activeTag = $element;
			}
		}
	}

	/**
	 * End element handler for XML parser
	 *
	 * If the end of <item>, <channel>, <image> or <textinput>
	 * is reached, this method updates the structure array
	 * $this->struct[] and adds the field "type" to this array,
	 * that defines the type of the current field.
	 *
	 * @access private
	 * @param  resource $parser XML parser object
	 * @param  string $element
	 * @return void
	 */
	function endHandler($parser, string $element): void{//@phan-suppress-current-line PhanUnusedPublicNoOverrideMethodParameter
		if(substr($element, 0, 4) === "RSS:"){
			$element = substr($element, 4);
		}

		if($element == end($this->insideTagStack)){
			array_pop($this->insideTagStack);

			//$this->struct[] = array_merge(['type' => strtolower($element)], $this->last);
		}

		switch($element){
			case 'ITEM':
				$this->items[] = $this->fields['item'];
				$this->fields['item'] = [];
				break;
			case 'IMAGE':
				$this->images[] = $this->fields['image'];
				$this->fields['image'] = [];
				break;

			case 'TEXTINPUT':
				$this->textinputs[] = $this->fields['textinput'];
				$this->fields['textinput'] = [];
				break;

			case 'ENCLOSURE':
				if(!isset($this->fields['item']['enclosures'])){
					$this->fields['item']['enclosures'] = [];
				}

				$this->fields['item']['enclosures'][] = array_change_key_case($this->attribs, CASE_LOWER);
				$this->attribs = [];
				break;
		}

		$this->activeTag = '';
	}

	/**
	 * Handler for character data
	 *
	 * @access private
	 * @param resource $parser XML parser object
	 * @param string $cdata CDATA
	 */
	function cdataHandler($parser, string $cdata): void{//@phan-suppress-current-line PhanUnusedPublicNoOverrideMethodParameter
		if(($tag = end($this->insideTagStack))){
			$tagName = strtolower($tag);
			$var = constant('self::' . strtoupper($tagName) . '_TAGS');

			if(in_array($this->activeTag, $var, true) || in_array($this->activeTag, self::MODULE_TAGS, true)){
				$this->_add($tagName, strtolower($this->activeTag), $cdata);
			}
		}
	}

	/**
	 * Default handler for XML parser
	 *
	 * @access private
	 * @param  resource $parser XML parser object
	 * @param  string $cdata CDATA
	 * @return void
	 */
	function defaultHandler($parser, string $cdata): void{//@phan-suppress-current-line PhanUnusedPublicNoOverrideMethodParameter
		return;
	}

	/**
	 * Add element to internal result sets
	 *
	 * @access private
	 * @param  string $type Name of the result set
	 * @param  string $field Fieldname
	 * @param  string $value Value
	 * @see    cdataHandler
	 */
	private function _add(string $type, string $field, string $value): void{
		if(empty($this->fields[$type]) || empty($this->fields[$type][$field])){
			$this->fields[$type][$field] = $value;
		} else {
			$this->fields[$type][$field] .= $value;
		}

		$this->last = $this->fields[$type];
	}

	/**
	 * Get general information about current channel
	 *
	 * This method returns an array containing the information
	 * that has been extracted from the <channel>-tag while parsing
	 * the RSS file.
	 *
	 * @access public
	 * @return array
	 */
	function getChannelInfo(): array{
		return (array) $this->channel;
	}

	/**
	 * Get items from RSS file
	 *
	 * This method returns an array containing the set of items
	 * that are provided by the RSS file.
	 *
	 * @access public
	 * @return array<string,string>
	 */
	function getItems(): array{
		return (array) $this->items;
	}

	/**
	 * Get images from RSS file
	 *
	 * This method returns an array containing the set of images
	 * that are provided by the RSS file.
	 *
	 * @access public
	 * @return array
	 */
	function getImages(): array{
		return (array) $this->images;
	}

	/**
	 * Get text input fields from RSS file
	 *
	 * @access public
	 * @return array
	 */
	function getTextinputs(): array{
		return (array) $this->textinputs;
	}

}
